"
 COPYRIGHT (c) 1997 by eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"


SimpleView subclass:#MenuPanel
	instanceVariableNames:'style adornment shadowView mapTime mustRearrange superMenu
		shortKeyInset selection items groupSizes receiver enableChannel'
	classVariableNames:'InitialSelectionQuerySignal DefaultStyle DefaultAdornment
		GroupDividerSize'
	poolDictionaries:''
	category:'Views-Menus'
!

Object subclass:#Item
	instanceVariableNames:'layout menuPanel subMenu adornment rawLabel enableChannel nameKey
		accessCharacterPosition value label'
	classVariableNames:'HorizontalInset VerticalInset LabelRightOffset ShortcutKeyOffset
		IndicatorOn IndicatorOff'
	poolDictionaries:''
	privateIn:MenuPanel
!

Object subclass:#Adornment
	instanceVariableNames:'indication accessCharacterPosition shortcutKey argument'
	classVariableNames:''
	poolDictionaries:''
	privateIn:MenuPanel::Item
!

!MenuPanel class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1997 by eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

!

documentation
"
    a menu panel used for both pull-down-menus and pop-up-menus.

    not yet finished MenuPanel class - this will eventually replace
    most of the MenuView and PopUpMenu stuff.
    (and hopefully be ST-80 compatible ...)

    [author:]
        Claus Atzkern

    [see also:]
        Menu
        MenuItem
"

!

examples

"
    start as PullDownMenu
                                                                                [exBegin]
    |top subView mview desc s1 s2 s3 img lbs labels|

    top := StandardSystemView new.

    mview := self in:top.

    labels := #( '\b foo' 'bar' 'baz' 'test' 'claus' ).
    mview level:2.
    mview verticalLayout:false.
    img := Image fromFile:'bitmaps/SBrowser.xbm'.
    lbs := Array with:'foo' with:'bar' with:img with:'baz' with:'\b test' with:'ludwig \t'.
    mview labels:lbs.
    mview shortcutKeyAt:2 put:#Cut.
    mview accessCharacterPositionAt:1 put:1.
    mview accessCharacterPositionAt:2 put:2.

    mview enabledAt:5 put:false.
    mview groupSizes:#( 2 2 ).
    s1 := self labels:labels.
    s1 accessCharacterPositionAt:1 put:1.
    s1 accessCharacterPositionAt:2 put:2.
    s1 groupSizes:#( 2 2 ).
    s2 := self labels:#( '1' nil '2' '-' '3' '=' '4' ' ' '5' ).
    s3 := self labels:lbs.

    s1 subMenuAt:2 put:s2.
    s1 subMenuAt:3 put:(self labels:lbs).
    s2 subMenuAt:3 put:s3.
    s3 subMenuAt:3 put:(self labels:labels).
    s3 shortcutKeyAt:3 put:$q.

    mview subMenuAt:1 put:s1.
    mview subMenuAt:4 put:(self labels:lbs).
    (mview subMenuAt:4) shortcutKeyAt:3 put:#Copy.
    s1 shortcutKeyAt:1 put:#Copy.
    s1 shortcutKeyAt:3 put:#Paste.

    mview subMenuAt:2 put:(self labels:labels).
    top extent:(400 @ (mview height)).
    top open.
                                                                                [exEnd]


    start as PopUpMenu
                                                                                [exBegin]
    |subView mview desc s1 s2 s3 img lbs labels|

    mview := self new.
    labels := #( '\b foo' 'bar' 'baz' ).
    mview level:2.

    img := Image fromFile:'bitmaps/SBrowser.xbm'.
    lbs := Array with:'foo' with:'bar' with:img with:'baz' with:'\b test'.
    mview labels:lbs.

    s1 := self labels:labels.
    s2 := self labels:#( '1' nil '2' '-' '3' '=' '4' ' ' '5' ).
    s3 := self labels:lbs.
    s1 subMenuAt:2 put:s2.
    s1 subMenuAt:3 put:(self labels:lbs).
    s2 subMenuAt:3 put:s3.
    s3 subMenuAt:3 put:(self labels:labels).
    s3 shortcutKeyAt:3 put:$q.

    mview subMenuAt:1 put:s1.
    mview subMenuAt:4 put:(self labels:lbs).
    (mview subMenuAt:4) shortcutKeyAt:3 put:#Copy.
    s1 shortcutKeyAt:1 put:#Copy.
    s1 shortcutKeyAt:3 put:#Paste.

    mview subMenuAt:2 put:(self labels:labels).
    mview startUp
                                                                                [exEnd]


    start from menu spec
                                                                                [exBegin]
    |menu|

    menu := self menu:
        #(#Menu #( #(#MenuItem 
                    #label: 'File' 
                    #submenu:
                      #(#Menu #(#(#MenuItem #label: 'quit' #value:#quit )     
                                 (#MenuItem 
                                    #label: 'edit' 
                                    #submenu:
                                      #(#Menu #( #(#MenuItem #label: 'edit'  #value:#edit )     
                                                 #(#MenuItem #label: 'close' #value:#close)     
                                               )
                                               nil
                                               nil
                                       )     
                                  )
                                 #(#MenuItem #label: 'help' #value:#help )     
                               )
                               nil
                               nil
                       )     
                 ) 
                #(#MenuItem #label: 'Inspect' #value:#inspectMenu ) 
                #(#MenuItem #label: 'Bar' 
                            #submenu:
                               #(#Menu #( #(#MenuItem #label: 'bar 1' #value:#bar1 )     
                                          #(#MenuItem #label: 'bar 2' #value:#bar2 )     
                                        )
                                        nil
                                        nil
                                )     
                 ) 
              ) 
              #( 2 )
              nil
         ) decodeAsLiteralArray.  

    menu verticalLayout:false.
    Transcript showCR:(menu startUp).
                                                                                [exEnd]

"

! !

!MenuPanel class methodsFor:'instance creation'!

fromSpec:aSpec
    ^ self fromSpec:aSpec receiver:nil
!

fromSpec:aSpec receiver:aReceiver
    |menu|

    aSpec notNil ifTrue:[
        menu := Menu new.
        menu fromLiteralArrayEncoding:aSpec.
    ].
  ^ self menu:menu receiver:aReceiver
!

labels:labels
    ^ self labels:labels nameKeys:nil receiver:nil
!

labels:labels nameKeys:nameKeys
    ^ self labels:labels nameKeys:nameKeys receiver:nil
!

labels:labels nameKeys:nameKeys receiver:aReceiver
    |mview|

    mview := self menu:nil receiver:aReceiver.
    mview labels:labels.
    mview nameKeys:nameKeys.
  ^ mview

!

menu:aMenu
    ^ self menu:aMenu receiver:nil
!

menu:aMenu receiver:aReceiver
    |mview|

    mview := self new.
    mview menu:aMenu.

"/ a menu itself may contain a receiver
"/ thus we do not overwrite the receiver

    aReceiver notNil ifTrue:[
        mview receiver:aReceiver
    ].
  ^ mview

! !

!MenuPanel class methodsFor:'class initialization'!

initialize
    InitialSelectionQuerySignal isNil ifTrue:[
        InitialSelectionQuerySignal := QuerySignal new.
    ].

    DefaultStyle isNil ifTrue:[
        DefaultStyle := IdentityDictionary new.

        DefaultStyle at:#foregroundColor         put:(Color black).
        DefaultStyle at:#backgroundColor         put:DefaultViewBackgroundColor.
        DefaultStyle at:#activeBackgroundColor   put:(Color white).
        DefaultStyle at:#disabledForegroundColor put:(Color grey:25).
        DefaultStyle at:#groupDividerSize        put:3.
        DefaultStyle at:#fitFirstPanel           put:true.
        DefaultStyle at:#horizontalSpace         put:0.
        DefaultStyle at:#borderWidth             put:1.
    ].

    DefaultAdornment isNil ifTrue:[
        DefaultAdornment := IdentityDictionary new.

        DefaultAdornment at:#showSeparatingLines     put:false.
        DefaultAdornment at:#showGroupDivider        put:true.
        DefaultAdornment at:#verticalLayout          put:true.
        DefaultAdornment at:#item                    put:nil.
        DefaultAdornment at:#value                   put:nil.
        DefaultAdornment at:#menuHolder              put:nil.
        DefaultAdornment at:#enableChannel           put:nil.
    ].

"
DefaultStyle := nil.
DefaultAdornment := nil.
self initialize
"
! !

!MenuPanel class methodsFor:'private'!

subMenu:aSubMenu
    "create a submenu; can be redifined in derived classes
    "
  ^ (MenuPanel new) menu:aSubMenu.

! !

!MenuPanel class methodsFor:'tests'!

claus
"
self claus
"
    |menu panel|

    menu := 
        #(#Menu #( #(#MenuItem 
                    #label: 'File' 
                    #submenu:
                      #(#Menu #(#(#MenuItem #label: 'quit' #value:#quit )     
                                 (#MenuItem 
                                    #label: 'edit' 
                                    #submenu:
                                      #(#Menu #( #(#MenuItem #label: 'edit'  #value:#edit )     
                                                 #(#MenuItem #label: 'close window' #value:#close)     
                                               )
                                               nil
                                               nil
                                       )     
                                  )
                                 #(#MenuItem #label: 'help' #value:#help )     
                               )
                               nil
                               nil
                       )     
                 ) 
                #(#MenuItem #label: 'Inspect' #value:#inspectMenu ) 
                #(#MenuItem #label: 'Bar' 
                            #submenu:
                               #(#Menu #( #(#MenuItem #label: 'bar 1' #value:#bar1 )     
                                          #(#MenuItem #label: 'bar 2' #value:#bar2 )     
                                        )
                                        nil
                                        nil
                                )     
                 ) 
              ) 
              #( 2 )
              nil
         ) decodeAsLiteralArray.  

    panel := self menu:menu.
    panel verticalLayout:false.
    Transcript showCR:(panel startUp).

!

st80test
"
MenuPanel st80test
"
     |menu|

     menu := #(#Menu #( #(#MenuItem 
                #rawLabel: '&File' 
                #submenu: #(#Menu #(
                    #(#MenuItem 
                            #rawLabel: '&Save As...'
                            #indication: true
                            #value: #imageSaveAs ) 
                    #(#MenuItem 
                            #rawLabel: '&Perm Save As...' 
                            #value: #filePermSaveAs ) 
                    #(#MenuItem 
                            #rawLabel: 'Perm &Undo As...' 
                            #value: #filePermUndoAs ) 
                    #(#MenuItem 
                            #rawLabel: '&Collect Garbage' 
                            #value: #collectGarbage ) 
                    #(#MenuItem 
                            #rawLabel: 'Collect All &Garbage' 
                            #value: #collectAllGarbage ) 
                    #(#MenuItem 
                            #rawLabel: 'Se&ttings' 
                            #value: #visualWorksSettings ) 
                    #(#MenuItem 
                            #rawLabel: 'E&xit VisualWorks...' 
                            #value: #visualWorksExit ) 
                ) #(3 2 1 1 ) nil ) ) 
            #(#MenuItem 
                #rawLabel: '&Browse' 
                #submenu: #(#Menu #(
                    #(#MenuItem 
                            #rawLabel: '&All Classes' 
                            #value: #browseAllClasses 
                            #labelImage: #(#ResourceRetriever nil #allClassesIcon ) ) 
                    #(#MenuItem 
                            #rawLabel: 'Class &Named...' 
                            #value: #browseClassNamed ) 
                    #(#MenuItem 
                            #rawLabel: '&Resources' 
                            #value: #browseApplications 
                            #labelImage: #(#ResourceRetriever nil #finderIcon ) ) 
                    #(#MenuItem 
                            #rawLabel: 'References &To...' 
                            #value: #browseSendersOf ) 
                    #(#MenuItem 
                            #rawLabel: '&Implementors Of...' 
                            #value: #browseImplementorsOf )
                ) #(3 2 ) nil ) ) 
            #(#MenuItem 
                #rawLabel: '&Tools' 
                #submenu: #(#Menu #(
                    #(#MenuItem 
                            #rawLabel: '&File List' 
                            #value: #openFileList 
                            #labelImage: #(#ResourceRetriever nil #fileListIcon ) ) 
                    #(#MenuItem 
                            #rawLabel: 'File &Editor...' 
                            #value: #openFileEditor ) 
                    #(#MenuItem 
                            #rawLabel: '&Workspace' 
                            #value: #toolsNewWorkspace 
                            #labelImage: #(#ResourceRetriever nil #workspaceIcon ) ) 
                    #(#MenuItem 
                            #rawLabel: 'Parcel &List' 
                            #value: #openParcelList ) 
                    #(#MenuItem 
                            #rawLabel: '&New Canvas' 
                            #value: #toolsNewCanvas 
                            #labelImage: #(#ResourceRetriever nil #newCanvasIcon ) ) 
                    #(#MenuItem 
                            #rawLabel: '&Palette' 
                            #value: #toolsPalette ) 
                    #(#MenuItem 
                            #rawLabel: '&Canvas Tool' 
                            #value: #toolsCanvasTool ) 
                    #(#MenuItem 
                            #rawLabel: '&Image Editor' 
                            #value: #toolsMaskEditor ) 
                    #(#MenuItem 
                            #rawLabel: '&Menu Editor' 
                            #value: #toolsMenuEditor ) 
                    #(#MenuItem 
                            #rawLabel: '&Advanced' 
                            #nameKey: #advanced ) 
                    #(#MenuItem 
                            #rawLabel: '&DLL and C Connect' 
                            #nameKey: #dllcc 
                            #value: #openExternalFinder 
                            #labelImage: #(#ResourceRetriever nil #extFinderIcon ) ) 
                    #(#MenuItem 
                            #rawLabel: 'System &Transcript' 
                            #nameKey: #transcript 
                            #value: #toggleSystemTranscript 
                            #indication: true )
                ) #(4 5 2 1 ) nil ) ) 
            #(#MenuItem 
                #rawLabel: '&Changes' 
                #submenu: #(#Menu #(
                    #(#MenuItem 
                            #rawLabel: 'Open Change &List' 
                            #value: #changesOpenChangeList ) 
                    #(#MenuItem 
                            #rawLabel: '&File Out Changes...' 
                            #value: #changesFileOutChanges ) 
                    #(#MenuItem 
                            #rawLabel: '&Empty Changes...' 
                            #value: #changesEmptyChanges ) 
                    #(#MenuItem 
                            #rawLabel: 'Changed &Methods' 
                            #value: #changesChangedMethods ) 
                    #(#MenuItem 
                            #rawLabel: 'Inspect &ChangeSet' 
                            #value: #changesInspectChangeSet ) 
                    #(#MenuItem 
                            #rawLabel: 'Open &Project' 
                            #value: #changesOpenProject ) 
                    #(#MenuItem 
                            #rawLabel: 'E&xit Project' 
                            #value: #changesExitProject )
                ) #(1 4 2 ) nil ) ) 
            #(#MenuItem 
                #rawLabel: '&Database' 
                #submenu: #(#Menu #(
                    #(#MenuItem 
                            #rawLabel: 'Ad Hoc &SQL' 
                            #nameKey: #adHoc 
                            #value: #openAdHocQuery ) 
                    #(#MenuItem 
                            #rawLabel: 'Data &Modeler' 
                            #nameKey: #dataModeler 
                            #value: #openDataModelBrowser 
                            #labelImage: #(#ResourceRetriever nil #dbToolIcon ) ) 
                    #(#MenuItem 
                            #rawLabel: 'Canvas &Composer' 
                            #nameKey: #canvasComposer 
                            #value: #openCanvasComposer ) 
                    #(#MenuItem 
                            #rawLabel: 'New Data &Form...' 
                            #nameKey: #dataForm 
                            #value: #newDataForm ) 
                    #(#MenuItem 
                            #rawLabel: 'New Database &Application...' 
                            #nameKey: #dataBaseAp 
                                #value: #newDataMain )
                ) #(1 4 ) nil ) ) 
            #(#MenuItem 
                #rawLabel: '&Window' 
                #submenu: #(#Menu #(
                    #(#MenuItem 
                            #rawLabel: 'Re&fresh All' 
                            #value: #winRefreshAll ) 
                    #(#MenuItem 
                            #rawLabel: '&Collapse All' 
                            #value: #winCollapseAll ) 
                    #(#MenuItem 
                            #rawLabel: 'Re&store All' 
                            #value: #winRestoreAll ) 
                    #(#MenuItem 
                            #rawLabel: '&Windows' 
                            #nameKey: #windowsMenu )
                ) #(3 1 ) nil ) ) 
            #(#MenuItem 
                #rawLabel: '&Help' 
                #submenu: #(#Menu #(
                    #(#MenuItem 
                            #rawLabel: 'Open Online &Documentation' 
                            #value: #openHelpBrowser 
                            #labelImage: #(#ResourceRetriever nil #helpIcon ) ) 
                    #(#MenuItem 
                            #rawLabel: '&Quick Start Guides...' 
                            #value: #openGuidingDialog ) 
                    #(#MenuItem 
                            #rawLabel: 'About &VisualWorks...' 
                            #value: #helpAbout )
                ) #(2 1 ) nil ) ) 
      ) #(7 ) nil
    )
    decodeAsLiteralArray.

    Transcript showCR:(menu startUp).



!

test
"
self test
"
    |top subView mview desc s1 s2 s3 img lbs labels|

    top := StandardSystemView new.

    mview := self in:top.

    labels := #( 'foo' 'bar' 'baz' 'test' 'claus' ).
    mview level:2.
    mview verticalLayout:false.
    img := Image fromFile:'bitmaps/SBrowser.xbm'.
    lbs := Array with:'foo' with:'bar' with:img with:'baz' with:'test' with:'ludwig'.
    mview labels:lbs.
    mview shortcutKeyAt:2 put:#Cut.
    mview accessCharacterPositionAt:1 put:1.
    mview accessCharacterPositionAt:2 put:2.

    mview enabledAt:5 put:false.
    mview groupSizes:#( 2 2 ).
    s1 := self labels:labels.
    s1 accessCharacterPositionAt:1 put:1.
    s1 accessCharacterPositionAt:2 put:2.
    s1 groupSizes:#( 2 2 ).
    s2 := self labels:#( '1' nil '2' '-' '3' '=' '4' ' ' '5' ).
    s3 := self labels:lbs.

    s1 subMenuAt:2 put:s2.
    s1 subMenuAt:3 put:(self labels:lbs).
    s2 subMenuAt:3 put:s3.
    s3 subMenuAt:3 put:(self labels:labels).
    s3 shortcutKeyAt:3 put:$q.

    mview subMenuAt:1 put:s1.
    mview subMenuAt:4 put:(self labels:lbs).
    (mview subMenuAt:4) shortcutKeyAt:3 put:#Copy.
    s1 shortcutKeyAt:1 put:#Copy.
    s1 shortcutKeyAt:3 put:#Paste.

    mview subMenuAt:2 put:(self labels:labels).
    top extent:(400 @ (mview height)).
    top open.

! !

!MenuPanel methodsFor:'accept'!

accept
    "accept current selected item
    "
    self accept:(self selection)
!

accept:anItem
    "this is the topMenu: accept item
    "
    |value item topMenu|

    self superMenu notNil ifTrue:[
        ^ self topMenu accept:anItem
    ].
    self selection:nil.
    self ungrabMouseAndKeyboard.

    anItem notNil ifTrue:[
        value := anItem accept.

        value notNil ifTrue:[
            item := anItem
        ]
    ].

    self menuAdornmentAt:#value put:value.
    self menuAdornmentAt:#item  put:item.

    self isPopUpView ifTrue:[
        self destroy
    ] ifFalse:[
        self do:[:el| el updateIndicators].
    ].
  ^ item.

!

lastItemAccepted
    "returns last item selected or nil
    "
  ^ self topMenu menuAdornmentAt:#item
!

lastValueAccepted
    "returns last value accepted or nil
    "
  ^ self topMenu menuAdornmentAt:#value.
! !

!MenuPanel methodsFor:'accessing'!

accessCharacterPositionAt:stringOrNumber
    "get the access character position for a textLabel
    "
  ^ self itemAt:stringOrNumber do:[:el| el accessCharacterPosition ]
!

accessCharacterPositionAt:stringOrNumber put:anIndexOrNil
    "get the access character position for a textLabel
    "
    self itemAt:stringOrNumber do:[:el| el accessCharacterPosition:anIndexOrNil ]
!

accessCharacterPositions
    "returns a collection of accessCharacterPosition's or nil
    "
    ^ self collect:[:anItem| anItem accessCharacterPosition ]

!

accessCharacterPositions:something
    "define accessCharacterPosition's for each item
    "
    self onEachPerform:#accessCharacterPosition: withArgList:something
!

args
    "returns a collection of argument's or nil
    "
    ^ self collect:[:anItem| anItem argument ]
!

args:something
    "define arguments for each item
    "
    self onEachPerform:#argument: withArgList:something
!

argsAt:stringOrNumber
    "gets the argument of an item or nil
    "
  ^ self itemAt:stringOrNumber do:[:el| el argument ]
!

argsAt:stringOrNumber put:anArgument
    "sets the argument of an item
    "
    self itemAt:stringOrNumber do:[:el| el argument:anArgument ]
!

groupSizes
    "gets collection of group sizes
    "
  ^ groupSizes
!

groupSizes:aGroupSizes
    "sets collection of group sizes
    "
    aGroupSizes = groupSizes ifFalse:[
        groupSizes := aGroupSizes copy.
        self mustRearrange.
    ].
!

labelAt:stringOrNumber
    "gets the label of an item or nil
    "
  ^ self itemAt:stringOrNumber do:[:el| el label ]
!

labelAt:stringOrNumber put:aLabel
    "sets the label of an item
    "
    self itemAt:stringOrNumber do:[:el| el label:aLabel ]
!

labels
    "returns a collection of labels's or nil
    "
    ^ self collect:[:anItem| anItem label ]

!

labels:labels
    "define labels for each item
    "
    self disabledRedrawDo:[
        self removeAll.

        labels notNil ifTrue:[
            labels do:[:aLabel|(self createAtIndex:nil) label:aLabel]
        ]
    ]
!

nameKeyAt:stringOrNumber
    "gets the nameKey of an item or nil
    "
  ^ self itemAt:stringOrNumber do:[:el| el nameKey ]
!

nameKeyAt:stringOrNumber put:aNameKey
    "sets the nameKey of an item
    "
    self itemAt:stringOrNumber do:[:el| el nameKey:aNameKey ]
!

nameKeys
    "returns a collection of nameKeys's or nil
    "
    ^ self collect:[:anItem| anItem nameKey ]
!

nameKeys:something
    "define nameKeys for each item
    "
    self onEachPerform:#nameKey: withArgList:something
!

numberOfItems
    "gets number of items
    "
    ^ items size
!

receiver
    "get the menu-receiver. Thats the one who gets the messages ( both from myself and
     from all submenus no specific receiver is defined ).
    "
    (receiver isNil and:[superMenu notNil]) ifTrue:[
        ^ superMenu receiver
    ].
  ^ receiver
!

receiver:anObject 
    "set the menu-receiver. Thats the one who gets the messages ( both from myself and
     from all submenus no specific receiver is defined ).
    "
    receiver := anObject
!

shortcutKeyAt:stringOrNumber
    "gets the shortCutKey of an item or nil
    "
  ^ self itemAt:stringOrNumber do:[:el| el shortcutKey ]
!

shortcutKeyAt:stringOrNumber put:aKey
    "sets the shortCutKey of an item
    "
    self itemAt:stringOrNumber do:[:el| el shortcutKey:aKey ]
!

shortcutKeys
    "returns a collection of shortcutKey's or nil
    "
    ^ self collect:[:anItem| anItem shortcutKey ]

!

shortcutKeys:something
    "define shortcutKey's for each item
    "
    self onEachPerform:#shortcutKey: withArgList:something
!

valueAt:stringOrNumber
    "gets value of an item; a block, valueHolder, ...
    "
  ^ self itemAt:stringOrNumber do:[:el| el value ]

!

valueAt:stringOrNumber put:someThing
    "sets value of an item; a block, valueHolder, ...
    "
    self itemAt:stringOrNumber do:[:el| el value:someThing ]


!

values:something
    "define values for each item
    "
    self onEachPerform:#value: withArgList:something
! !

!MenuPanel methodsFor:'accessing behavior'!

disable
    "disable the menu and all its entries
    "
    self enabled:false
!

disableAll
    "disable all items
    "
    self do:[:anItem| anItem enabled:false ]
!

disableAll:collectionOfIndicesOrNames
    "disable an collection of items
    "
    collectionOfIndicesOrNames do:[:entry| self enabledAt:entry put:false ].

!

enable
    "enable the menu and all its entries
    "
    self enabled:true
!

enableAll
    "enable all items
    "
    self do:[:anItem| anItem enabled:true ]
!

enableAll:collectionOfIndicesOrNames
    "enable an collection of items
    "
    collectionOfIndicesOrNames do:[:entry| self enabledAt:entry put:true ].

!

enabled
    "returns enabled state
    "
    superMenu isNil ifTrue:[
        ^ (self findFirst:[:anItem|anItem enabled]) ~~ 0
    ].
  ^ superMenu enabledAt:(superMenu findFirst:[:anItem|anItem submenu == self])
!

enabled:aState
    "change enabled state of menu
    "
    |idx state|

    state := aState ? true.

    superMenu isNil ifTrue:[
        self do:[:anItem| anItem enabled:state ]
    ] ifFalse:[
        idx := superMenu findFirst:[:anItem| anItem submenu == self ].
        superMenu enabledAt:idx put:state.
    ]
!

enabledAt:stringOrNumber
    "gets the enabled state of an item or false
    "
  ^ self itemAt:stringOrNumber do:[:el| el enabled ] ifAbsent:false
!

enabledAt:stringOrNumber put:aState
    "sets the enabled state of an item
    "
    self itemAt:stringOrNumber do:[:el| el enabled:aState ]
!

isEnabled:stringOrNumber
    "gets the enabled state of an item or false
    "
    ^ self enabledAt:stringOrNumber
! !

!MenuPanel methodsFor:'accessing channels'!

enableChannel
    "gets a enable channel or nil
    "
    adornment isNil ifTrue:[^ nil].
  ^ adornment at:#enableChannel ifAbsent:nil
!

enableChannel:aValueHolder
    "set my enableChannel
    "
    |enableChannel|

    (enableChannel := self enableChannel) notNil ifTrue:[
        enableChannel removeDependent:self
    ].

    aValueHolder notNil ifTrue:[
        adornment isNil ifTrue:[
            adornment := DefaultAdornment copy
        ].
        adornment at:#enableChannel put:aValueHolder.
        aValueHolder addDependent:self.
    ].
    self enabled:(aValueHolder value)
!

menuHolder
    "gets a menu holder or nil
    "
    adornment isNil ifTrue:[^ nil].
  ^ adornment at:#menuHolder ifAbsent:nil
!

menuHolder:aValueHolder
    "set a menu holder
    "
    |menuHolder|

    (menuHolder := self menuHolder) notNil ifTrue:[
        menuHolder removeDependent:self
    ].

    aValueHolder notNil ifTrue:[
        adornment isNil ifTrue:[
            adornment := DefaultAdornment copy
        ].
        adornment at:#menuHolder put:aValueHolder.
        aValueHolder addDependent:self.
    ].
    self menu:(aValueHolder value)
! !

!MenuPanel methodsFor:'accessing color & font'!

activeBackgroundColor
    "return the background color used to highlight selections
    "
  ^ self styleAt:#activeBackgroundColor
!

activeBackgroundColor:aColor
    "set the background drawing color used to highlight selection. You should not 
     use this method; instead leave the value as defined in the styleSheet.
    "
    self styleAt:#activeBackgroundColor put:aColor
!

backgroundColor
    "return the background color
    "
    ^ self styleAt:#backgroundColor
!

backgroundColor:aColor
    "set the background drawing color. You should not use this method;
     instead leave the value as defined in the styleSheet.
    "
    self styleAt:#backgroundColor put:aColor
!

disabledForegroundColor
    "return the foreground color used by disabled items
    "
  ^ self styleAt:#disabledForegroundColor

!

disabledForegroundColor:aColor
    "set the foregroundColor drawing color used by disabled items. You should not
     use this method; instead leave the value as defined in the styleSheet.
    "
    self styleAt:#disabledForegroundColor put:aColor
!

font:aFont
    "set the font
    "
    superMenu notNil ifTrue:[
        self topMenu font:aFont
    ] ifFalse:[
        (aFont notNil and:[aFont ~~ font]) ifTrue:[
            super font:(aFont on:device).
            self mustRearrange.
        ]
    ]


!

foregroundColor
    "return the passive foreground color
    "
  ^ self styleAt:#foregroundColor


!

foregroundColor:aColor
    "set the foregroundColor drawing color. You should not use this method;
     instead leave the value as defined in the styleSheet.
    "
    self styleAt:#foregroundColor put:aColor
!

lightColor
    "get the lightColor
    "
    ^ lightColor


!

lightColor:aColor
    "set the light drawing color. You should not use this method;
     instead leave the value as defined in the styleSheet.
    "
    lightColor ~~ aColor ifTrue:[
        super lightColor:aColor.

        shown ifTrue:[
            self invalidate 
        ]
    ]


!

shadowColor
    "get the shadowColor
    "
    ^ shadowColor


!

shadowColor:aColor
    "set the shadow drawing color. You should not use this method;
     instead leave the value as defined in the styleSheet.
    "
    shadowColor ~~ aColor ifTrue:[
        super shadowColor:aColor.

        shown ifTrue:[
            self invalidate 
        ]
    ]


! !

!MenuPanel methodsFor:'accessing dimensions'!

height
    "default height
    "
    |item|

    (explicitExtent ~~ true) ifTrue:[
        (item := self itemAt:1) notNil ifTrue:[
            self rearrangeItems.
          ^ item height
        ].
        ^ 4 + (font height + (font descent * 2)).
    ].
    ^ super height
!

preferredExtent
    "compute and returns my preferred extent
    "
    |x y hasMenu shCtKey|

    self numberOfItems == 0 ifTrue:[
        ^ 32 @ 32
    ].
    self isFitPanel ifTrue:[
        x := 0
    ] ifFalse:[
        x := groupSizes size * self groupDividerSize.
    ].

    self verticalLayout ifFalse:[
        y := 0.

        self do:[:el||z|
            x := x + el preferredExtentX.
            y < (z := el preferredExtentY) ifTrue:[y := z]
        ].
        x := x + ((items size - 1) * self horizontalSpace).
    ] ifTrue:[
        hasMenu := false.
        shCtKey := 0.
        y := x.
        x := 0.

        self do:[:el| |l e|
            (l := el rawLabel) notNil ifTrue:[
                (e := l widthOn:self) > x ifTrue:[x := e].

                el hasSubmenu ifTrue:[hasMenu := true].

                (     (l := el shortcutKeyAsString) notNil
                 and:[(e := l widthOn:self) > shCtKey]
                ) ifTrue:[
                    shCtKey := e
                ].
            ].
            y := y + el preferredExtentY
        ].
        x := x + Item horizontalInset.

        (hasMenu or:[shCtKey ~~ 0]) ifTrue:[
            x := x + Item labelRightOffset + shCtKey + ((items at:1) rightArrow width).

            (shCtKey ~~ 0 and:[hasMenu]) ifTrue:[
                x := x + (Item shortcutKeyOffset) 
            ]
        ].
        x := x + Item horizontalInset.
    ].
  ^ x @ y
        

!

shortKeyInset
    "left inset of shortcutKey
    "
  ^ shortKeyInset
! !

!MenuPanel methodsFor:'accessing items'!

itemAt:stringOrNumber
    "returns item assigned to an index, nameKey, textLabel or value if symbol.
     If no item match nil is returned.
    "
    |idx|

    idx := self indexOf:stringOrNumber.
    (idx > 0 and:[idx <= items size]) ifTrue:[ ^ items at:idx ].
  ^ nil

!

itemAt:stringOrNumber do:aOneArgBlock
    "evaluate teh block for an item and return the result from the block. In case that  
     the item not exists nil is returned
    "
    ^ self itemAt:stringOrNumber do:aOneArgBlock ifAbsent:nil
!

itemAt:stringOrNumber do:aOneArgBlock ifAbsent:exceptionBlock
    "evaluate teh block for an item and return the result from the block. In case that  
     the item not exists the result of the exception block is returned (no arguments).
    "
    |item|

    item := self itemAt:stringOrNumber.
    item notNil ifTrue:[ ^ aOneArgBlock value:item ].
  ^ exceptionBlock value
! !

!MenuPanel methodsFor:'accessing look'!

fitFirstPanel
    "gets true if the first panel in the menu hierarchy must be fit 
     to the extent of its superView
    "
  ^ self styleAt:#fitFirstPanel
!

fitFirstPanel:aState
    "sets true if the first panel in the menu hierarchy must be fit 
     to the extent of its superView
    "
    self styleAt:#fitFirstPanel put:aState
!

groupDividerSize
    "get the size of the group dividers
    "
  ^ self styleAt:#groupDividerSize
!

groupDividerSize:aSize
    "set the size of the group dividers
    "
    self styleAt:#groupDividerSize put:aSize
!

horizontalSpace
    "get the horizontal space between to items
    "
  ^ self styleAt:#horizontalSpace
!

horizontalSpace:aSize
    "set the horizontal space between to items
    "
    self styleAt:#horizontalSpace put:aSize
!

showGroupDivider
    "get the enabled flag for showing groupDiveders
    "
  ^ self menuAdornmentAt:#showGroupDivider
!

showGroupDivider:aState
    "set the enabled flag for showing groupDiveders
    "
    self menuAdornmentAt:#showGroupDivider put:aState
!

showSeparatingLines
    "gets true if drawing of separating lines is enabled.
    "
  ^ self menuAdornmentAt:#showSeparatingLines
!

showSeparatingLines:aState
    "turn on/off drawing of separating lines.
    "
    self menuAdornmentAt:#showSeparatingLines put:aState
!

verticalLayout
    "get the layout: or vertical( true ) or horizontal( false )
    "
  ^ self menuAdornmentAt:#verticalLayout
!

verticalLayout:aState
    "set the layout: or vertical( true ) or horizontal( false )
    "
    self menuAdornmentAt:#verticalLayout put:aState.
! !

!MenuPanel methodsFor:'accessing submenu'!

subMenuAt:stringOrNumber
    "gets the submenu of an item or nil
    "
  ^ self itemAt:stringOrNumber do:[:el| el submenu ]
!

subMenuAt:stringOrNumber put:aSubMenu
    "sets the submenu of an item
    "
    self itemAt:stringOrNumber do:[:el| el submenu:aSubMenu ]

!

subMenuShown
    "return the currently visible submenu - or nil if there is none
    "
    |item|

    (item := self selection) notNil ifTrue:[
        ^ item submenu
    ].
  ^ nil
! !

!MenuPanel methodsFor:'activation'!

hide
    "hide the view, leave its modal event loop
    "

    self selection:nil.
    self unmap.


!

show
    "realize the view at its last position
    "
  ^ self showAt:(self origin) resizing:true

!

showAt:aPoint
    "realize the view at aPoint
    "
  ^ self showAt:aPoint resizing:true


!

showAt:aPoint resizing:aBoolean
    "realize the view at aPoint
    "
    self rearrangeItems.

    aBoolean ifTrue:[
        self fixSize.
    ].
    self origin:aPoint.
    self makeFullyVisible.
    self openModal:[true]. "realize     "

  ^ self lastValueAccepted

!

showAtPointer
    "realize the view at the current pointer position
    "
  ^ self showAt:(device pointerPosition) resizing:true


!

showCenteredIn:aView
    "make myself visible at the screen center.
    "
    |top|

    top := aView topView.
    top raise.
  ^ self showAt:(top origin + (aView originRelativeTo:top) + (aView extent // 2) - (self extent // 2))


!

startUp
    "realize the menu at the current pointer position
    "
    ^ self showAtPointer
! !

!MenuPanel methodsFor:'adding & removing'!

createAtIndex:anIndexOrNil
    "create an item and add this item to the index. In case of nil the item
     is added to the end. If the index is not valid nil is returned otherwise
     the new created item.
    "
    |max item|

    max := (items size) + 1.

    anIndexOrNil notNil ifTrue:[
        (anIndexOrNil < 1 or:[anIndexOrNil > max]) ifTrue:[
            ^ nil
        ]
    ].
    items isNil ifTrue:[
        items := OrderedCollection new
    ].
    item := Item in:self.

    (anIndexOrNil isNil or:[anIndexOrNil == max]) ifTrue:[
        items add:item
    ] ifFalse:[
        items add:item beforeIndex:anIndexOrNil
    ].
    ^ item
!

remove:stringOrNumber
    "remove the first item which is assigned to stringOrNumber;
     if found, remove and return it
    "
    |item|

    (item := self itemAt:stringOrNumber) notNil ifTrue:[
        items remove:item.
        item  destroy.
        items isEmpty ifTrue:[items := nil].
        self mustRearrange.
    ].
  ^ item
!

removeAll
    "remove all items and submenus
    "
    self disabledRedrawDo:[
        self selection:nil.
        groupSizes := nil.
        self do:[:el| el destroy ].
        items := nil
    ].
! !

!MenuPanel methodsFor:'change & update'!

update:something with:aParameter from:changedObject
    |channel|

    channel := self menuHolder.
    channel == changedObject ifTrue:[^ self menu:(channel value)].

    channel := self enableChannel.
    channel == changedObject ifTrue:[^ self enabled:(channel value)].

    super update:something with:aParameter from:changedObject
! !

!MenuPanel methodsFor:'converting'!

asMenu
    "convert contents to menu
    "
    |menu|

    menu := Menu new.
    menu groupSizes:groupSizes.
    self do:[:anItem| menu addItem:(anItem asMenuItem) ].
  ^ menu
!

fromSpec:aMenuSpec
    "build from spec
    "
    |menu|

    menu := Menu new.
    menu fromLiteralArrayEncoding:aMenuSpec.
    self menu:menu

!

menu:aMenu
    "convert to Menu
    "
    |menu|

    self disabledRedrawDo:[
        self removeAll.

        (menu := aMenu) notNil ifTrue:[
            (aMenu isCollection) ifTrue:[
                menu := Menu new.
                menu fromLiteralArrayEncoding:aMenu.
            ] ifFalse:[
                menu receiver notNil ifTrue:[
                    receiver := menu receiver
                ]
            ].
            menu menuItems notNil ifTrue:[
                menu menuItems do:[:anItem|
                    (self createAtIndex:nil) menuItem:anItem
                ]
            ].
            self groupSizes:(menu groupSizes).
        ]
    ]    

! !

!MenuPanel methodsFor:'drawing'!

disabledRedrawDo:aBlock
    "evaluate a block without redrawing within the block; after processing
     of the block a redraw might be performed
    "
    |state|

    state := mustRearrange.
    mustRearrange := true.
    aBlock value.
    mustRearrange := state.
    self mustRearrange
!

mustRearrange
    "returns true if layout of items must be recomputed
    "
    mustRearrange ifFalse:[
        mustRearrange := true.
        self invalidate.
    ]
!

rearrangeItems
    "recompute layout of my items
    "
    |expLast x y e grpDivSz hrzSpace layout|

    mustRearrange ifFalse:[ ^ self ].

"/  fetch font from superMenu
    (superMenu notNil and:[superMenu font ~~ font]) ifTrue:[
        super font:(superMenu font on:device)
    ].

    items isNil ifTrue:[
        mustRearrange := false.
      ^ self
    ].
    hrzSpace := self horizontalSpace.
    expLast  := false.

    self hasGroupDividers ifTrue:[
        self isFitPanel ifFalse:[
            grpDivSz := self groupDividerSize
        ] ifTrue:[
            expLast := true.
            x := 0.
            e := self computeExtent.

            self verticalLayout ifTrue:[
                items do:[:el|x := x + el preferredExtentY].
                y := e y.
            ] ifFalse:[
                items do:[:el|x := x + el preferredExtentX].
                x := x + ((items size - 1) * hrzSpace).
                y := e x.
            ].
            (grpDivSz := (y - x) // (groupSizes size)) <= 0 ifTrue:[
                grpDivSz := nil
            ]
        ]
    ].

    (self isPopUpView or:[explicitExtent ~~ true]) ifTrue:[
        e := self preferredExtent.

        self isPopUpView ifFalse:[
            self verticalLayout ifTrue:[e y:1.0]
                               ifFalse:[e x:1.0]
        ].
        self extent:e
    ] ifFalse:[
        e := self computeExtent
    ].

    x := 0.
    y := 0.

    self keysAndValuesDo:[:anIndex :el||org|
        org := Point x:x y:y.

        self verticalLayout ifTrue:[
            y := y + el preferredExtentY.
            el layout:(Rectangle origin:org corner:(e x @ y)).
        ] ifFalse:[
            x := x + el preferredExtentX.
            el layout:(Rectangle origin:org corner:(x @ e y)).
            x := x + hrzSpace.
        ].

        (grpDivSz notNil and:[self hasGroupDividerAt:anIndex]) ifTrue:[
            self verticalLayout ifTrue:[y := y + grpDivSz]
                               ifFalse:[x := x + grpDivSz]
        ]
    ].
    expLast ifTrue:[
        layout := items last layout.

        self verticalLayout ifTrue:[layout bottom:((self extent y) + 1)]
                           ifFalse:[layout right:((self extent x) + 1)].
    ].
    mustRearrange := false.





!

redrawX:x y:y width:w height:h
    "redraw a rectangle
    "
    |start end isVrt x1 x2 y1 y2 item layout lnSz|

    shown ifFalse:[^ self].

    super redrawX:x y:y width:w height:h.
    self  paint:(self backgroundColor).
    self  clearRectangleX:x y:y width:w height:h.

    isVrt := self verticalLayout.
    end   := items size.

    mustRearrange ifTrue:[
        self isPopUpView not ifTrue:[
            explicitExtent := true
        ].
        self rearrangeItems.
        start := 1
    ] ifFalse:[
        end == 0 ifTrue:[ ^ self ].

        isVrt ifTrue:[
            start := self findFirst:[:el| (el layout bottom) >= y ].
            start == 0 ifTrue:[ ^ self ].
            end := y + h.
            end := self findLast:[:el| (el layout top) < end ].
        ] ifFalse:[
            start := self findFirst:[:el| (el layout right) >= x ].
            start == 0 ifTrue:[ ^ self ].
            end := x + w.
            end := self findLast:[:el| (el layout left) < end ].
        ].

        (start ~~ 1 and:[self hasGroupDividerAt:(start-1)]) ifTrue:[
            start := start - 1
        ]
    ].

    (     self hasGroupDividers
     and:[self showGroupDivider
     and:[self isFitPanel not]]
    ) ifTrue:[
        lnSz := self groupDividerSize // 2
    ].

    start to:end do:[:i|
        item := items at:i.
        item redraw.

        (lnSz notNil and:[self hasGroupDividerAt:i]) ifTrue:[
            layout := item layout.

            isVrt ifTrue:[
                x1 := layout left.
                x2 := layout right.
                y1 := (layout bottom) + lnSz.
                y2 := y1.
            ] ifFalse:[
                x1 := (layout right) + lnSz.
                x2 := x1.
                y1 := layout top.
                y2 := layout bottom.
            ].
            self paint:(self shadowColor).
            self displayLineFromX:x1 y:y1 toX:x2 y:y2.
            self paint:(self lightColor).

            isVrt ifTrue:[y1 := y1 + 1. y2 := y1 ]
                 ifFalse:[x1 := x1 + 1. x2 := x1 ].

            self displayLineFromX:x1 y:y1 toX:x2 y:y2
        ]
    ]


! !

!MenuPanel methodsFor:'enumerting & searching'!

collect:aOneArgBlock
    "evaluate the argument, aOneArgBlock for every item in the menuPanel
     and return a collection of the results
    "
    items notNil ifTrue:[^ items collect:aOneArgBlock ].
  ^ nil

!

do:aOneArgBlock
    "evaluate the argument, aOneArgBlock for every item in the menuPanel.
    "
    items notNil ifTrue:[ items do:aOneArgBlock ].

!

findFirst:aOneArgBlock
    "find the first item, for which evaluation of the argument, aOneArgBlock
     returns true; return its index or 0 if none detected.
    "
    items notNil ifTrue:[ ^ items findFirst:aOneArgBlock ].
  ^ 0

!

findLast:aOneArgBlock
    "find the last item, for which evaluation of the argument, aOneArgBlock
     returns true; return its index or 0 if none detected.
    "
    items notNil ifTrue:[ ^ items findLast:aOneArgBlock ].
  ^ 0

!

indexOf:something
    "returns index of an item assigned to an index, nameKey, textLabel or value if symbol.
     If no item match 0 is returned. No range checks are performed on a number argument
    "
    |i v|

    something isNumber ifTrue:[ ^ something ].
    something isNil    ifTrue:[ ^ 0 ].

    i := self findFirst:[:el|el nameKey = something].

    i ~~ 0 ifTrue:[
        ^ i
    ].

    something isSymbol ifTrue:[
        i := self findFirst:[:el|
            v := el value.
            v isSymbol and:[v == something]
        ].
        i ~~ 0 ifTrue:[
            ^ i
        ]
    ].

    (something respondsTo:#string) ifTrue:[
        v := something string.
      ^ self findFirst:[:el|el textLabel = v].
    ].
  ^ 0
!

keysAndValuesDo:aTwoArgBlock
    "evaluate the argument, aTwoArgBlock for every item in the menuPanel.
    "
    items notNil ifTrue:[ items keysAndValuesDo:aTwoArgBlock ].

! !

!MenuPanel methodsFor:'event handling'!

buttonMotion:state x:x y:y
    "open or close the corresponding submenus
    "
    |menu point sensor|

    (    (sensor := self sensor) notNil
     and:[sensor hasButtonMotionEventFor:nil]
    ) ifTrue:[
        ^ self
    ].

    (self containsPointX:x y:y) ifTrue:[
        ^ self selection:(self itemAtX:x y:y)
    ].
    menu := self superMenuAtX:x y:y.

    menu isNil ifTrue:[
        ^ self selection:nil
    ].
    point := self translatePoint:(x@y) to:menu.
    menu selection:(menu itemAtX:(point x) y:(point y))
!

buttonPress:button x:x y:y
    "any button pressed; open or close the corresponding submenus
    "
    |menu point item|

    menu := self superMenuAtX:x y:y.

    menu isNil ifTrue:[
        menu := self topMenu.
    ] ifFalse:[
        point := self translatePoint:(x@y) to:menu.
        item  := menu itemAtX:(point x) y:(point y)
    ].
    menu selection:item
!

buttonRelease:button x:x y:y
    "button release action; accept selection and close all views
    "
    ( self hasSelection or:[
        (OperatingSystem millisecondTimeDeltaBetween:(Time millisecondClockValue)
                                    and:mapTime) > 200]
    ) ifTrue:[
        self accept
    ]
!

cursorPressed:aKey
    "handle a cursor key
    "
    |idx next first submenu item|

    (self hasSelection not and:[superMenu notNil]) ifTrue:[
        ^ superMenu cursorPressed:aKey
    ].

    self verticalLayout ifTrue:[
        aKey == #CursorLeft  ifTrue:[^ self selection:nil].
        aKey ~~ #CursorRight ifTrue:[next := aKey == #CursorDown].
    ] ifFalse:[
        aKey == #CursorUp ifTrue:[^ self selection:nil].
        aKey ~~ #CursorDown ifTrue:[next := aKey == #CursorRight].        
    ].

    next isNil ifTrue:[
        (item := self selection) notNil ifTrue:[
            (submenu := item submenu) notNil ifTrue:[
                idx := submenu findFirst:[:el| el canSelect ].
              ^ submenu selectionIndex:idx
            ].
          ^ self selection:nil
        ].
      ^ self
    ].

    first := self findFirst:[:el| el canSelect ].
    first == 0 ifTrue:[^ self].

    idx := self selectionIndex.

    next ifTrue:[
        [((idx := idx + 1) <= items size and:[(items at:idx) canSelect not])
        ] whileTrue.

        idx > items size ifTrue:[
            idx := first
        ].
    ] ifFalse:[    
        [((idx := idx - 1) > 0  and:[(items at:idx) canSelect not])
        ] whileTrue.
        
        idx < 1 ifTrue:[ idx := self findLast:[:el| el canSelect ] ]
    ].
    self selectionIndex:idx

!

keyPress:key x:x y:y
    "any key is pressed
    "
    |listOfItems item menu idx cIdx upperKey lowerKey|

    menu := self.

    [ menu shown ] whileFalse:[
        (menu := superMenu) isNil ifTrue:[^ self]
    ].

    key == #Return ifTrue:[
        ^ menu accept
    ].

    (     key == #CursorDown or:[key == #CursorUp
      or:[key == #CursorLeft or:[key == #CursorRight]]]
    ) ifTrue:[
        ^ menu cursorPressed:key
    ].

    listOfItems := self selectItemsForShortcutKey:key.
    listOfItems notNil ifTrue:[
        item := listOfItems first.

        item hasSubmenu ifFalse:[
            ^ item superMenu accept:item
        ].
      ^ self openMenusFromItems:listOfItems
    ].

    (self hasSelection not and:[superMenu notNil]) ifTrue:[
        (superMenu containsPoint:(self translatePoint:(x@y) to:superMenu)) ifTrue:[
            menu := superMenu
        ]
    ].

    cIdx := menu selectionIndex.
    cIdx isNil ifTrue:[cIdx := 0].

    upperKey := key asUppercase.
    lowerKey := key asLowercase.

    menu keysAndValuesDo:[:anIndex :el| |c l|
        (el canSelect and:[(l := el textLabel) notNil]) ifTrue:[
            l size ~~ 0 ifTrue:[
                (c := el accessCharacter) notNil ifTrue:[
                    (c == upperKey or:[c == lowerKey]) ifTrue:[
                        ^ menu selection:el
                    ]
                ] ifFalse:[
                    ((c := l first) == upperKey or:[c == lowerKey]) ifTrue:[
                        anIndex > cIdx ifTrue:[
                            ^ menu selection:el
                        ] ifFalse:[
                            idx isNil ifTrue:[
                                idx := anIndex
                            ] ifFalse:[
                                anIndex > idx ifTrue:[
                                    anIndex ~~ cIdx ifTrue:[
                                        idx := anIndex
                                    ]
                                ]
                            ]
                        ]
                        
                    ]
                ]
            ]
        ]
    ].
    (item := menu itemAt:idx) isNil ifTrue:[
        menu hasSelection ifFalse:[
            (menu := menu superMenu) isNil ifTrue:[
                ^ super keyPress:key x:x y:y
            ]
        ]
    ].
    menu selection:item.

!

sizeChanged:how
    self isFitPanel ifTrue:[
        self mustRearrange.
    ].
    super sizeChanged:how
! !

!MenuPanel methodsFor:'grabbing'!

grabMouseAndKeyboard
    "get exclusive access to pointer and keyboard
    "
    |sensor|

    realized ifTrue:[
        sensor := self sensor.

        device activePointerGrab ~~ self ifTrue:[
            sensor notNil ifTrue:[
                sensor flushMotionEventsFor:nil.
            ].

            (device grabPointerInView:self) ifFalse:[
                Delay waitForSeconds:0.1.
                (device grabPointerInView:self) ifFalse:[
                    "give up"
                    'PopUpView [warning]: could not grab pointer' errorPrintCR.
                    self unmap
                ]
            ]
        ].

        device activeKeyboardGrab ~~ self ifTrue:[
            sensor notNil ifTrue:[
                device sync.
                sensor flushKeyboardFor:nil
            ].
            device grabKeyboardInView:self.
            self getKeyboardFocus
        ]
    ]


!

ungrabMouseAndKeyboard
    "ungrap resources( mouse and keyboard )
    "
    |sensor|

    device activePointerGrab == self ifTrue:[
        device ungrabPointer.
    ].
    device activeKeyboardGrab == self ifTrue:[
        (sensor := self sensor) notNil ifTrue:[
            "/ make certain all X events have been received
            device sync.
            "/ now all events have been received.
            "/ now, flush all pointer events
            sensor flushKeyboardFor:self
        ].
        device ungrabKeyboard.
    ].
    (superMenu notNil and:[superMenu shown and:[superMenu isPopUpView]]) ifTrue:[
        superMenu grabMouseAndKeyboard
    ]
        


! !

!MenuPanel methodsFor:'initialize / release'!

create
    "create the shadow view for a none contained submenu; see initStyle
    "
    super create.

    self isPopUpView ifTrue:[
        shadowView isNil ifTrue:[
            shadowView := (ShadowView onDevice:device) for:self
        ]
    ] ifFalse:[
        explicitExtent == true ifTrue:[
            (self extent x) == (superView extent x) ifTrue:[
                self verticalLayout:false
            ]
        ]
    ]
!

destroy
    "destroy items and shadowView; remove dependencies
    "
    |channel|

    self do:[:el|el destroy].

    (channel := self menuHolder)    notNil ifTrue:[ channel removeDependent:self].
    (channel := self enableChannel) notNil ifTrue:[ channel removeDependent:self].

    super destroy.
    superMenu := nil.
    items     := nil.
    shadowView notNil ifTrue:[shadowView destroy].

!

initialize
    "set default configuration
    "
    super initialize.
    font := font on:device.
    self origin:0.0@0.0.
    originChanged  := extentChanged := false.
    explicitExtent := nil.
    shortKeyInset  := 0.
    mustRearrange  := false.
    self level:1.
!

mapped
    "grab the pointer here, when visible (but not control is already lost). 
     If the grab fails, try again and unmap myself if that fails too.
    "
    |anItemList|

    mapTime := Time millisecondClockValue.
    super mapped.

    anItemList := InitialSelectionQuerySignal raise.

    self isPopUpView ifTrue:[
        self grabMouseAndKeyboard
    ] ifFalse:[
        super viewBackground:(self backgroundColor)
.   ].
    self do:[:el| el updateIndicators ].

    anItemList size > 0 ifTrue:[
        self redrawX:0 y:0 width:10000 height:10000.
        self openMenusFromItems:anItemList.
    ].

!

realize
    "realize menu and shadowView
    "
    |bsz|

    self isPopUpView ifTrue:[
        shadowView notNil ifTrue:[
            bsz := self styleAt:#borderWidth.
            bsz ~~ self borderWidth ifTrue:[
                self borderWidth:bsz
            ].
            shadowView realize
        ].
        self raise.
    ].
    super realize.


!

recreate
    "this is called after a snapin. If the image was saved with an active menu,
     hide the menu
    "
    self selection:nil.
    super recreate.

!

unmap
    "unmap view
    "
    self ungrabMouseAndKeyboard.
    super unmap.
    shadowView notNil ifTrue:[shadowView unmap].


! !

!MenuPanel methodsFor:'misc'!

superMenu
    "returns supermenu or nil
    "
    ^ superMenu


!

superMenu:aSuperMenu
    "set the supermenu starting from
    "
    superMenu := aSuperMenu.
!

topMenu
    "returns the topMenu; the one having no superMenu
    "
    |top|

    top := self.

    [ top superMenu notNil ] whileTrue:[
        top := top superMenu
    ].
  ^ top


! !

!MenuPanel methodsFor:'private'!

menuAdornmentAt:aSymbol
    "returns a value derived from adornment
    "
    |adm|

    adm := adornment ? DefaultAdornment.
  ^ adm at:aSymbol ifAbsent:nil
!

menuAdornmentAt:aSymbol put:something
    "sets a value for the specific menu
    "
    |adm|

    (self menuAdornmentAt:aSymbol) ~~ something ifTrue:[
        adornment isNil ifTrue:[
            adornment := DefaultAdornment copy
        ].
        adornment at:aSymbol put:something.
        self mustRearrange.
    ]
!

onEachPerform:aSelector withArgList:aList
    "on each item perform selector with an argument derived from aList
    "
    |size|

    aList isCollection ifTrue:[
        items size >= aList size ifTrue:[
            aList keysAndValuesDo:[:anIndex :anArg|
                (items at:anIndex) perform:aSelector with:anArg
            ]
        ]
    ] ifFalse:[
        self do:[:anItem| anItem perform:aSelector with:aList ]
    ]


!

openMenusFromItems:anItemList
    "open all menus derived from sequence of items
    "
    |item|

    (anItemList isNil or:[anItemList isEmpty]) ifTrue:[
        ^ self
    ].
    item := anItemList removeLast.

    item enabled ifFalse:[
        ^ self
    ].

    InitialSelectionQuerySignal answer:anItemList do:[
        self selection:item
    ]
!

selectItemsForShortcutKey:aKey
    "get sequence of items up to the item providing the key (inclusive). The
     first entry into the collection is the item providing the key, the last
     entry is the item in the topMenu( reverse )
    "
    |seq|

    self do:[:anItem|
        anItem isEnabled ifTrue:[
            anItem shortcutKey = aKey ifTrue:[
                seq := OrderedCollection new.
            ] ifFalse:[
                anItem hasSubmenu ifTrue:[
                    seq := anItem submenu selectItemsForShortcutKey:aKey
                ]
            ].
            seq notNil ifTrue:[
                seq add:anItem.
              ^ seq
            ]
        ]
    ].
  ^ nil
        
!

styleAt:aSymbol
    "gets a value derived from the style in the topMenu
    "
    |aStyle|

    superMenu isNil ifTrue:[
        aStyle := style ? DefaultStyle.
      ^ aStyle at:aSymbol ifAbsent:nil
    ].
  ^ superMenu styleAt:aSymbol

!

styleAt:aSymbol put:aValue
    "sets a value derived from the style in the topMenu
    "
    |aStyle|

    superMenu isNil ifTrue:[
        aStyle := style ? DefaultStyle.

        (aStyle at:aSymbol ifAbsent:nil) ~~ aValue ifTrue:[
            style isNil ifTrue:[
                style := DefaultStyle copy.
            ].
            style at:aSymbol put:aValue.
        ]
    ] ifFalse:[
        superMenu styleAt:aSymbol put:aValue
    ]
!

translatePoint:aPoint to:aView
    "translate a point into a views point; in case of no view nil is returned
    "
    aView notNil ifTrue:[
        aView == self ifTrue:[^ aPoint].
      ^ device translatePoint:aPoint from:(self id) to:(aView id)
    ].
  ^ nil

! !

!MenuPanel methodsFor:'private searching'!

itemAtX:x y:y
    "returns item at a point or nil
    "
    self do:[:el| (el containsPointX:x y:y) ifTrue:[^el] ].
  ^ nil


!

superMenuAtX:x y:y
    "returns supermenu at a point or nil
    "
    |menu|

    (self containsPointX:x y:y) ifTrue:[^ self].
    menu := self.

    [ (menu := menu superMenu) notNil ] whileTrue:[
        (menu containsPoint:(self translatePoint:(x@y) to:menu)) ifTrue:[
            ^ menu
        ]
    ].
  ^ nil


! !

!MenuPanel methodsFor:'queries'!

canDrawItem
    "returns true if an item could be drawn otherwise false
    "
    ^ (mustRearrange not and:[self shown])


!

containsPoint:aPoint
    "returns true if point is contained by the view
    "
    ^ self containsPointX:(aPoint x) y:(aPoint y)
!

containsPointX:x y:y
    "returns true if point is contained by the view
    "
    |ext|

    (x >= 0 and:[y >= 0]) ifTrue:[
        ext := self computeExtent.
      ^ (x < ext x and:[y < ext y])
    ].
    ^ false


!

hasGroupDividerAt:anIndex
    "returns true if a divider is defined at an index
    "
    |i|

    groupSizes size ~~ 0 ifTrue:[
        i := 0.

        groupSizes do:[:t|
            (i := i + t) == anIndex ifTrue:[
                ^ true
            ]
        ]
    ].
  ^ false

!

hasGroupDividers
    "returns true if any group divider exists
    "
  ^ (items size ~~ 0 and:[groupSizes size ~~ 0])
!

isEnabled
    "returns enabled state of menu and items
    "
    ^ self enabled
!

isFitPanel
    "returns true if the panel is the first in the menu hierarchy in must
     be fit to the extent of its superView
    "
    superView isNil ifTrue:[
        ^ false
    ].
  ^ self styleAt:#fitFirstPanel
!

isPopUpView
    "return true if view is a popup view; without decoration
     and popUp to top immediately
    "
    ^ superView isNil


!

isVerticalLayout
    "returns true if vertical layout otherwise false( horizontal layout )
    "
  ^ self verticalLayout


!

type
    "return the views type. This is one of #normal,
     #master, #slave or #partner."

     superView notNil ifTrue:[
        ^ superView type
     ].
   ^ #slave

! !

!MenuPanel methodsFor:'selection'!

hasSelection
    "returns true if a selection exists
    "
    ^ self selection notNil
!

isValidSelection:something
    "returns true if something could be selected
    "
    |item|

    (item := self itemAt:something) notNil ifTrue:[
        ^ item canSelect
    ].
  ^ false
!

selection
    "returns current selected item or nil
    "
    ^ selection


!

selection:anItemOrNil
    "change selection to an item
    "
    |item newSel|

    selection isNumber ifTrue:[
        newSel := self itemAt:anItemOrNil
    ] ifFalse:[
        (anItemOrNil notNil and:[anItemOrNil canSelect]) ifTrue:[
            newSel := anItemOrNil
        ]
    ].

    selection == newSel ifTrue:[^ self].

    (item := selection) notNil ifTrue:[
        selection := nil.
        item selected:false.
    ].
    newSel isNil ifTrue:[
        self isPopUpView ifFalse:[
            self ungrabMouseAndKeyboard
        ].
        ^ self
    ].
    selection := newSel.
    selection selected:true
!

selectionIndex
    "returns index of current selection or 0
    "
    |item|

    (item := self selection) notNil ifTrue:[
        ^ self findFirst:[:el| el == item ]
    ].
    ^ 0

!

selectionIndex:anIndex
    "set selection at an index
    "
    self selection:(self itemAt:anIndex)
! !

!MenuPanel::Item class methodsFor:'accessing'!

horizontalInset
    ^ HorizontalInset
!

labelRightOffset
    ^ LabelRightOffset
!

shortcutKeyOffset
    ^ ShortcutKeyOffset
!

verticalInset
    ^ VerticalInset
! !

!MenuPanel::Item class methodsFor:'init style'!

initStyle
    "setup defaults
    "
    HorizontalInset   := 4.
    VerticalInset     := 4.
    LabelRightOffset  := 15.
    ShortcutKeyOffset := 5.

"/    IndicatorOn  := Image fromFile:'RadioOn.xbm'.
"/    IndicatorOff := Image fromFile:'RadioOff.xbm'.

    IndicatorOn  := CheckToggle checkImageForStyle:#borderedFatcross.
    IndicatorOff := CheckToggle checkImageForStyle:#border.

"
self initStyle
"
! !

!MenuPanel::Item class methodsFor:'instance creation'!

in:aSuperMenu
    ^ self in:aSuperMenu label:nil
!

in:aSuperMenu label:aLabel
    |item|

    HorizontalInset isNil ifTrue:[ self initStyle ].

    item := self new in:aSuperMenu.
    item label:aLabel.
  ^ item
!

in:aSuperMenu menuItem:aMenuItem
    |item|

    item := self in:aSuperMenu.
    item menuItem:aMenuItem.
  ^ item.
! !

!MenuPanel::Item methodsFor:'accept'!

accept
    "item accepted; perform action, ...
    "
    |sel res rcv num arg tval|

    (self enabled and:[self hasSubmenu not]) ifFalse:[
        ^ nil
    ].

    self hasIndication ifTrue:[
        tval := self indicationValue not.
        self indicationValue:tval
    ].

    value isNil ifTrue:[
        ^ menuPanel findFirst:[:el| el == self ]
    ].

    (arg := self argument) isNil ifTrue:[
        arg := tval
    ].

    value isSymbol ifFalse:[
        rcv := value.
        (rcv respondsTo:#numArgs) ifFalse:[
            ^ rcv
        ].
        num := value numArgs.
        res := menuPanel findFirst:[:el| el == self ].

        num == 0 ifTrue:[
            sel := #value
        ] ifFalse:[
            num == 1 ifTrue:[sel := #value:]
                    ifFalse:[sel := #value:value:]
        ]
    ] ifTrue:[
        (rcv := menuPanel receiver) isNil ifTrue:[
            ^ value
        ].
        sel := value.
        num := value numArgs.
        res := value.
    ].

    num == 0 ifTrue:[
        rcv perform:sel
    ] ifFalse:[
        num == 1 ifTrue:[
            rcv perform:sel with:arg
        ] ifFalse:[
            rcv perform:sel with:arg with:self
        ]
    ].
  ^ res
! !

!MenuPanel::Item methodsFor:'accessing'!

accessCharacter
    "returns my accessCharacter or nil
    "
    accessCharacterPosition isNil ifTrue:[
        ^ nil
    ].
  ^ (rawLabel string) at:accessCharacterPosition
!

accessCharacterPosition
    "get the access character position or nil
    "
  ^ accessCharacterPosition
!

accessCharacterPosition:anIndex
    "set the access character position or nil
    "
    |idx lbl|

    (      (accessCharacterPosition ~~ anIndex)
      and:[(lbl := self textLabel) notNil]
    ) ifTrue:[
        anIndex notNil ifTrue:[
            (anIndex < 1 or:[anIndex > lbl size]) ifTrue:[
                ^ self
            ].
        ].
        accessCharacterPosition := anIndex.
        self updateRawLabel.
    ].
!

argument
    "gets the argument
    "
    adornment isNil ifTrue:[^ nil ].
  ^ adornment argument
!

argument:anArgument
    "sets the argument
    "
    self argument ~~ anArgument ifTrue:[
        self adornment argument:anArgument.
    ]
!

compareAccessCharacterWith:aKey
    "returns true if key is my access character
    "
    |s|

    accessCharacterPosition notNil ifTrue:[
        s := (rawLabel string) at:accessCharacterPosition.
        s == aKey ifTrue:[^ true ].
    ].
  ^ false
!

label
    "returns the label
    "
    ^ label
!

label:aLabel
    "set a new label; if the label changed, a redraw is performed;
     handle characters $& (ST-80 compatibility)
    "
    |type form i rest s|

    accessCharacterPosition := nil.
    label := aLabel value.

    (label isString and:[(s := label size) > 1]) ifTrue:[
        i := 1.

        [((i := label indexOf:$& startingAt:i) ~~ 0 and:[i < s])] whileTrue:[
            rest := label copyFrom:(i+1).

            i == 1 ifTrue:[label := rest]
                  ifFalse:[label := (label copyFrom:1 to:(i-1)), rest].

            (label at:i) == $& ifTrue:[i := i + 1]
                              ifFalse:[accessCharacterPosition := i].
            s := s - 1.
        ]
    ].

    self updateRawLabel
!

menuPanel
    "returns my menuPanel
    "
    ^ menuPanel
!

nameKey
    "gets the nameKey
    "
    ^ nameKey
!

nameKey:aNameKey
    "sets the nameKey
    "
    nameKey := aNameKey.
!

rawLabel
    "returns my printable Label
    "
    ^ rawLabel
!

shortcutKey
    "get the key to press to select the submenu from the keyboard or if
     no submenu exists evaluate the action assigned to the item (accept).
    "
    adornment isNil ifTrue:[^ nil ].
  ^ adornment shortcutKey
!

shortcutKey:aKey
    "set the key to press to select the submenu from the keyboard or if
     no submenu exists evaluate the action assigned to the item (accept).
    "
    self shortcutKey ~~ aKey ifTrue:[
        self adornment shortcutKey:aKey.
        self redraw.
    ].
!

submenu
    "returns my submenu or nil
    "
    ^ subMenu
!

submenu:aSubMenu
    "set a new submenu; an existing submenu will be destroyed. This might lead
     to a redraw if 'hasSubmenu' changed
    "
    |hadMenu|

    (hadMenu := subMenu notNil) ifTrue:[
        subMenu destroy
    ].

    (subMenu := aSubMenu) notNil ifTrue:[
        aSubMenu class == Menu ifTrue:[
            subMenu := menuPanel class subMenu:aSubMenu
        ].
        subMenu superMenu:menuPanel
    ].

    hadMenu ~~ (subMenu notNil) ifTrue:[
        self redraw
    ]

!

textLabel
    "returns my textLabel or nil if none text
    "
    (rawLabel respondsTo:#string) ifTrue:[
        ^ rawLabel string
    ].
  ^ nil
!

value
    "gets value
    "
    ^ value
!

value:something
    "could be a value holder, an action or selector
    "
    value := something.
! !

!MenuPanel::Item methodsFor:'accessing behavior'!

enabled
    "returns the enabled state
    "
    ^ enableChannel value ? true
!

enabled:something
    "change the enabled state; if the state changed, a redraw is performed
    "
    |oldState newState|

    enableChannel isNil ifTrue:[
        oldState := true
    ] ifFalse:[
        oldState := enableChannel value.
        (self isKindOfValueHolder:enableChannel) ifTrue:[
            enableChannel removeDependent:self
        ]
    ].
    enableChannel := something.

    enableChannel isNil ifTrue:[
        newState := true
    ] ifFalse:[
        (self isKindOfValueHolder:enableChannel) ifTrue:[
            enableChannel addDependent:self
        ].
        newState := enableChannel value.
    ].

    newState ~~ oldState ifTrue:[
        self redraw
    ]

!

indication
    "get on/off indication
    "
    adornment isNil ifTrue:[^ nil].
  ^ adornment indication
!

indication:something
    "set on/off indication
    "
    |old|

    old := self indication.
    old == something ifTrue:[^ self].

    (self isKindOfValueHolder:old) ifTrue:[
        old removeDependent:self
    ].

    (self isKindOfValueHolder:something) ifTrue:[
        something addDependent:self
    ].
    self adornment indication:something.
    self updateRawLabel.
! !

!MenuPanel::Item methodsFor:'accessing dimension'!

height
    "gets height
    "
    layout isNil ifTrue:[
        ^ self preferredExtentY
    ].
  ^ layout height
!

layout
    "returns my layout ( Rectangle )
    "
    ^ layout
!

layout:aLayout
    "set a new layout ( Rectangle )
    "
    layout := aLayout.
    self redraw.
!

preferredExtentX
    "compute my preferred extent x
    "
    |x s|

    x := HorizontalInset + HorizontalInset.

    rawLabel notNil ifTrue:[
        x := x + (rawLabel widthOn:menuPanel).

        (s := self shortcutKeyAsString) notNil ifTrue:[
            x := x + LabelRightOffset + (s widthOn:menuPanel)
        ].
        menuPanel verticalLayout ifTrue:[
            self hasSubmenu ifTrue:[
                x := x + self rightArrow width.

                s notNil ifTrue:[
                    x := x + ShortcutKeyOffset
                ] ifFalse:[
                    x := x + LabelRightOffset
                ]
            ]
        ]
    ].
    ^ x
!

preferredExtentY
    "compute my preferred extent y
    "
    |y|

    rawLabel notNil ifTrue:[y := rawLabel heightOn:menuPanel]
                   ifFalse:[y := menuPanel font height].

  ^ y + VerticalInset + VerticalInset.
!

width
    "gets width
    "
    layout isNil ifTrue:[
        ^ self preferredExtentX
    ].
  ^ layout width
! !

!MenuPanel::Item methodsFor:'change & update'!

update:something with:aParameter from:changedObject
    changedObject == self indication ifTrue:[
        rawLabel icon:(self indicator)
    ].
    self redraw

!

updateIndicators
    "update indicators
    "
    |indicator| 

    (indicator := self indicator) notNil ifTrue:[
        indicator = rawLabel icon ifFalse:[
            rawLabel icon:indicator.
            self redraw.
        ]
    ].


! !

!MenuPanel::Item methodsFor:'constants'!

rightArrow
    "arrow used for submenu indication
    "
    ^ SelectionInListView rightArrowFormOn:(menuPanel device)
!

rightArrowShadow
    "shadow arrow used for submenu indication
    "
  ^ SelectionInListView rightArrowShadowFormOn:(menuPanel device)
! !

!MenuPanel::Item methodsFor:'converting'!

asMenuItem
    "convert to a MenuItem
    "
    |item|

    item := MenuItem labeled:(self label).
    item enabled:(self enabled).
    item accessCharacterPosition:(self accessCharacterPosition).
    item nameKey:(self nameKey).
    item shortcutKeyCharacter:(self shortcutKey).
    item value:(value value).
    item indication:(self indication value).

    subMenu notNil ifTrue:[
        item submenu:(subMenu asMenu)
    ].
  ^ item

!

menuItem:aMenuItem
    "setup attributes from a MenuItem
    "
    |acp lbl|

    menuPanel disabledRedrawDo:[
        label := nil.
        self enabled:(aMenuItem enabled).
        self nameKey:(aMenuItem nameKey).
        self indication:(aMenuItem indication).

        (lbl := aMenuItem labelImage value) isNil ifTrue:[
            lbl := aMenuItem label.
        ].
        self label:lbl.

        self shortcutKey:(aMenuItem shortcutKeyCharacter).

        (acp := aMenuItem accessCharacterPosition) notNil ifTrue:[
            self accessCharacterPosition:acp.
        ].

        self submenu:(aMenuItem submenu).
        self value:(aMenuItem value).
    ]
! !

!MenuPanel::Item methodsFor:'drawing'!

drawLabel
    "draw label
    "
    |y|

    self enabled ifTrue:[
        menuPanel paint:(menuPanel foregroundColor)
    ] ifFalse:[
        menuPanel paint:(menuPanel disabledForegroundColor)
    ].

    y := (layout top) + (((layout height) - (rawLabel heightOn:menuPanel)) // 2).

    (self textLabel) notNil ifTrue:[
        y := y + menuPanel font ascent.
    ].
    rawLabel displayOn:menuPanel x:((layout left)+HorizontalInset) y:y.
    self drawShortcutKey.
!

drawSeparatingLines
    "draw separating lines
    "
    |l r t b|

    menuPanel showSeparatingLines ifTrue:[
        l := layout left.
        r := layout right.
        t := layout top.
        b := layout bottom.

        menuPanel paint:(menuPanel lightColor).

        menuPanel verticalLayout ifTrue:[
            b := b - 1.
            t := t - 1.
            menuPanel displayLineFromX:l y:b toX:r y:b.
            menuPanel displayLineFromX:l y:t toX:r y:t.
            b := b - 1.
            t := t - 1.
            menuPanel paint:(menuPanel shadowColor).
            menuPanel displayLineFromX:l y:b toX:r y:b.
            menuPanel displayLineFromX:l y:t toX:r y:t.
        ] ifFalse:[
            r := r - 1.
            l := l - 1.
            menuPanel displayLineFromX:r y:t toX:r y:b.
            menuPanel displayLineFromX:l y:t toX:l y:b.
            r := r - 1.
            l := l - 1.
            menuPanel paint:(menuPanel shadowColor).
            menuPanel displayLineFromX:r y:t toX:r y:b.
            menuPanel displayLineFromX:l y:t toX:l y:b.
        ]
    ]
!

drawSeparator
    "draw line separator
    "
    |type l r y|

    type := self seperatorType.

    (type notNil and:[type ~~ #blankLine and:[menuPanel verticalLayout]]) ifTrue:[
        l := (layout left)  + HorizontalInset.
        r := (layout right) - HorizontalInset.
        y := (layout top) - 1 + ((layout height) // 2).

        type == #doubleLine ifTrue:[
            y := y - 2.
        ].
        menuPanel paint:(menuPanel shadowColor).
        menuPanel displayLineFromX:l y:y toX:r y:y.
        menuPanel paint:(menuPanel lightColor).
        y := y + 1.
        menuPanel displayLineFromX:l y:y toX:r y:y.

        type == #doubleLine ifTrue:[
            y := y + 3.
            menuPanel paint:(menuPanel shadowColor).
            menuPanel displayLineFromX:l y:y toX:r y:y.
            menuPanel paint:(menuPanel lightColor).
            y := y + 1.
            menuPanel displayLineFromX:l y:y toX:r y:y.
        ]
    ]

!

drawShortcutKey
    "draw the shorcutKey
    "
    |string x y|

    (string:= self shortcutKeyAsString) notNil ifTrue:[
        (x := menuPanel shortKeyInset) == 0 ifTrue:[
            x := LabelRightOffset + (rawLabel widthOn:menuPanel)
        ].
        x := (layout left) + x.
        y := (layout top) + (((layout height) - (string heightOn:menuPanel)) // 2).
        y := y + menuPanel font ascent.
        string displayOn:menuPanel x:x y:y.
    ]
!

drawSubmenuIndication
    "draw an arrow in case of having a submenu
    "
    |form x y col|

    (subMenu notNil and:[menuPanel verticalLayout]) ifTrue:[
        form := self rightArrow.
        x    := (layout right) - form width - HorizontalInset.
        y    := layout top + ((layout height - (form heightOn:menuPanel)) // 2).

        self isSelected ifFalse:[
            menuPanel paint:(menuPanel shadowColor).
            col := menuPanel lightColor.
        ] ifTrue:[
            menuPanel paint:(menuPanel lightColor).
            col := menuPanel shadowColor.
        ].
        menuPanel displayForm:form x:x y:y.
        menuPanel paint:col.
        menuPanel displayForm:(self rightArrowShadow) x:x y:y.
    ]

!

redraw
    "redraw item
    "
    |ownBgCol x y paint|

    menuPanel canDrawItem ifFalse:[
        ^ self
    ].
    self isSelected ifFalse:[
        paint := menuPanel backgroundColor
    ] ifTrue:[
        paint := menuPanel activeBackgroundColor
    ].

    (ownBgCol := self backgroundColorFromLabel) isNil ifTrue:[
        menuPanel paint:paint.
        menuPanel fillRectangle:layout.
    ] ifFalse:[
        self hasIndication ifFalse:[
            menuPanel paint:ownBgCol.
            menuPanel fillRectangle:layout.
        ] ifTrue:[
            menuPanel paint:paint.
            x := (rawLabel icon width) + HorizontalInset + 4.

            menuPanel fillRectangleX:(layout left)
                                   y:(layout top)
                               width:x
                              height:(layout height).

            menuPanel paint:ownBgCol.

            menuPanel fillRectangleX:(layout left + x)
                                   y:(layout top)
                               width:(layout width - x)
                              height:(layout height).
            ownBgCol := nil.
       ]
    ].
    self drawSeparatingLines.

    rawLabel notNil ifTrue:[
        self drawLabel.
        self drawSubmenuIndication.

        (ownBgCol notNil and:[self isSelected]) ifTrue:[
            x := Color black.
            y := Color white.

            ownBgCol brightness > 0.5 ifTrue:[menuPanel paint:(Color black)]
                                     ifFalse:[menuPanel paint:(Color white)].

            menuPanel displayRectangleX:(layout left   + 1)
                                      y:(layout top    + 1)
                                  width:(layout width  - 2)
                                 height:(layout height - 2).

            menuPanel displayRectangleX:(layout left   + 2)
                                      y:(layout top    + 2)
                                  width:(layout width  - 4)
                                 height:(layout height - 4).
        ]
    ] ifFalse:[
        self drawSeparator
    ]
! !

!MenuPanel::Item methodsFor:'initialization'!

destroy
    "destroy submenu
    "
    self submenu:nil.           "/ destroy submenus
    self enabled:nil.           "/ remove dependencies
    self indication:nil.        "/ remove dependencies
    menuPanel := nil.
!

in:aPanel
    "create item in a menuPanel
    "
    menuPanel := aPanel.
! !

!MenuPanel::Item methodsFor:'private'!

adornment
    "returns adornment; if not existing yet a new instance
     is created
    "
    adornment isNil ifTrue:[
        adornment := Adornment new
    ].
  ^ adornment
!

backgroundColorFromLabel
    "returns the background color derived from label or nil
    "
    |run|

    label isText ifFalse:[^ nil ].
    run := label emphasis.
    run size == 0 ifTrue:[^ nil ].

    run := run first.

    run size == 0 ifTrue:[
        (run value isColor and:[run key == #backgroundColor]) ifTrue:[
            ^ run value
        ]
    ] ifFalse:[
        run do:[:r|
            (r value isColor and:[r key == #backgroundColor]) ifTrue:[
                ^ r value
            ]
        ]
    ].
  ^ nil
!

indicationValue
    "returns indication value or nil
    "
    |indication val rcv sel|

    (indication := self indication) isNil ifTrue:[
        ^ nil
    ].

    indication isSymbol ifFalse:[
        val := indication value
    ] ifTrue:[
        rcv := menuPanel receiver.
        rcv notNil ifTrue:[
            (indication last) == $: ifFalse:[
                sel := indication
            ] ifTrue:[
                sel := indication copyFrom:1 to:((indication size) - 1)
            ].
            val := rcv perform:(sel asSymbol).
        ]
    ].
  ^ val == true
!

indicationValue:aValue
    "returns indication value or nil
    "
    |indication rcv|

    (indication := self indication) notNil ifTrue:[
        indication isSymbol ifTrue:[
            indication last == $: ifTrue:[
                (rcv := menuPanel receiver) notNil ifTrue:[
                    rcv perform:indication with:aValue
                ]
            ]
        ] ifFalse:[
            (self isKindOfValueHolder:indication) ifTrue:[
                indication value:aValue
            ]
        ]
    ]
!

indicator
    "returns indication form or nil
    "
    |value|

    (value := self indicationValue) isNil ifTrue:[
        ^ nil
    ].
  ^ value ifTrue:[IndicatorOn] ifFalse:[IndicatorOff]
!

seperatorType
    "returns type of seperator line or nil
    "
    |c lbl|

    rawLabel isNil ifFalse:[
        ^ nil
    ].

    (lbl := label value) isNil ifTrue:[
        ^ #singleLine
    ].

    lbl size == 1 ifTrue:[
        c := lbl first.
        c == $- ifTrue:[^ #singleLine].
        c == $= ifTrue:[^ #doubleLine].
    ].
  ^ #blankLine
!

updateRawLabel
    "recreate rawLabel
    "
    |char size indicator|

    (rawLabel := label value) isString ifFalse:[
        ^ menuPanel mustRearrange.
    ].

    rawLabel isText ifFalse:[
        rawLabel := rawLabel withoutSeparators
    ].
    size := rawLabel size.

    (indicator := self indicator) isNil ifTrue:[
        size == 0 ifTrue:[
            ^ menuPanel mustRearrange
        ].
            
        size == 1 ifTrue:[
            char := rawLabel first.

            (char == $- or:[char == $=]) ifTrue:[   "/ other line separators
                label := String new:1.
                label at:1 put:char.
                rawLabel := nil.
              ^ menuPanel mustRearrange
            ]
        ]
    ] ifFalse:[
        size == 0 ifTrue:[rawLabel := label value]
    ].

    rawLabel isText ifFalse:[
        rawLabel := Text string:rawLabel emphasis:#bold
    ].
    size := self accessCharacterPosition.

    (size notNil and:[size <= rawLabel size]) ifTrue:[
        rawLabel emphasisAt:size add:#underline
    ].

    (indicator := self indicator) notNil ifTrue:[
        rawLabel := LabelAndIcon icon:indicator string:rawLabel.
    ].
    menuPanel mustRearrange.
! !

!MenuPanel::Item methodsFor:'queries'!

canSelect
    "returns true if item is selectable
    "
    ^ (self enabled and:[rawLabel notNil])
!

containsPointX:x y:y
    "returns true if point is contained in my layout
    "
    ^ (     (x >= layout left)
        and:[x <  layout right
        and:[y >  layout top
        and:[y <= layout bottom]]]
      )
!

hasIndication
    "returns true if on/off indication exists
    "
  ^ self indication notNil
!

hasSubmenu
    "returns true if a submenu exists
    "
    ^ self submenu notNil
!

isEnabled
    "returns enabled state
    "
    ^ self enabled
!

isKindOfValueHolder:something
    "returns true if something is kind of vlaue holder
    "
    ^ ((something respondsTo:#value:) and:[something isBlock not])





!

shortcutKeyAsString
    "converts shortcutKey to a text object
    "
    |nm x key|

    "/ this is somewhat complicated: we have the symbolic key at hand,
    "/ but want to show the untranslated (inverse keyBoardMapped) key & modifier
    "/

    (key := self shortcutKey) isNil ifTrue:[
        ^ nil
    ].

    key isCharacter ifTrue:[
        nm := key asString
    ] ifFalse:[
        nm := menuPanel device keyboardMap keyAtValue:key ifAbsent:key.
        "/
        "/ some modifier-key combination ?
        "/

        (nm startsWith:#Cmd) ifTrue:[
            nm := (self shortcutKeyPrefixFor:#Cmd) , (nm copyFrom:4)
        ] ifFalse:[(nm startsWith:#Alt) ifTrue:[
            nm := (self shortcutKeyPrefixFor:#Alt) , (nm copyFrom:4)
        ] ifFalse:[(nm startsWith:#Meta) ifTrue:[
            nm := (self shortcutKeyPrefixFor:#Meta), (nm copyFrom:5)
        ]
        ifFalse:[(nm startsWith:#Ctrl) ifTrue:[
            nm := (self shortcutKeyPrefixFor:#Ctrl), (nm copyFrom:5)
        ] ifFalse:[
            nm := nm asString
        ]]]]
    ].
    ^ Text string:nm emphasis:#bold


!

shortcutKeyPrefixFor:aModifier
    "returns prefix assigned to a modifier
    "
    |m|

    m := menuPanel device modifierKeyTopFor:aModifier.

    m notNil ifTrue:[
        ^ m , '-'
    ].
    ^ aModifier.
! !

!MenuPanel::Item methodsFor:'selection'!

hideSubmenu
    "hide submenu
    "
    |windowGrp|

    subMenu realized ifFalse:[
        menuPanel device unmapWindow:subMenu id
    ] ifTrue:[
       subMenu hide
    ]
!

isSelected
    "returns true if item is selected
    "
    ^ menuPanel selection == self
!

openSubmenuAt:aPoint
    "open submenu at a point
    "
    |windowGrp|

    windowGrp := menuPanel topMenu windowGroup.

    windowGrp notNil ifTrue:[
        subMenu windowGroup:windowGrp.
        windowGrp addTopView:subMenu.
    ].
    subMenu fixSize.
    subMenu origin:aPoint.
    subMenu makeFullyVisible.

    subMenu realized ifFalse:[
        subMenu realize. 
    ] ifTrue:[
        menuPanel topMenu device mapWindow:subMenu id.
    ].

!

selected:aState
    "change selection to a state. Dependant on the state open or hide an existing
     submenu and perform a redraw
    "
    |p d subMenu|

    subMenu := self submenu.

    aState ifFalse:[
        self redraw.
        subMenu notNil ifTrue:[
            self hideSubmenu
        ].
      ^ self
    ].
    menuPanel shown ifFalse:[^ self].
    self redraw.

    subMenu isNil ifTrue:[
        ^ menuPanel grabMouseAndKeyboard
    ].

    menuPanel verticalLayout ifTrue:[p := (layout right) @ (layout top)]
                            ifFalse:[p := (layout left)  @ (layout bottom)].

    d := menuPanel device.
    p := d translatePoint:p from:(menuPanel id) to:(d rootWindowId).
    self openSubmenuAt:p.


! !

!MenuPanel::Item::Adornment methodsFor:'accessing'!

accessCharacterPosition
    "get the index of the access character in the label text or string, or nil if none
    "
    ^ accessCharacterPosition
!

accessCharacterPosition:anIndexOrNil
    "set the index of the access character in the label text or string, or nil if none
    "
    accessCharacterPosition := anIndexOrNil
!

argument
    "ST/X goody; get argunment to a selector or block
    "
  ^ argument
!

argument:anArgumentOrNil
    "ST/X goody; set argunment to a selector or block
    "
    argument := anArgumentOrNil
!

indication
    "get has on/off indicator value
    "
  ^ indication
!

indication:something
    "set has on/off indicator value
    "
    indication := something.
!

shortcutKey
    "get the character that is used as a shortcut key for this item
    "
  ^ shortcutKey
!

shortcutKey:aKeyOrNil
    "set the character that is used as a shortcut key for this item
    "
    shortcutKey := aKeyOrNil
! !

!MenuPanel class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libwidg2/MenuPanel.st,v 1.9 1997-06-11 13:56:53 ca Exp $'
! !
MenuPanel initialize!
